/* embbin.h
 * Embedded binaries' tables
 *
 * *** Do not edit this file directly, it was generated by genembbin.c
 * *** genembbin is not invoked neither compiled by make,
 * *** thus it should be compiled and executed manually.
 *
 * Ziproxy - the HTTP acceleration proxy
 * This code is under the following conditions:
 *
 * ---------------------------------------------------------------------
 * Copyright (c)2005-2021 Daniel Mealha Cabrita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 * ---------------------------------------------------------------------
 */


//To stop multiple inclusions.
#ifndef EMBBIN_H
#define EMBBIN_H

#define embbin_empty_image_size 49

const char embbin_empty_image [49] = {
	71, 73, 70, 56, 57, 97, 1, 0, 1, 0, -95, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 33, -7, 4, 1, 10, 0, 1, 0, 44, 0, 0, 0, 0, 1, 0, 1, 0, 0, 2, 2, 76, 1, 0, 59
};


#endif

